@@ -421,6 +421,7 @@ static int ext2_direct_scan(const char *device)
 	ext2fs_inode_bitmap inode_dir_map;
 	uid_t uid;
 	gid_t gid;
+	int ret = -1;
 
 	if ((error = ext2fs_open(device, 0, 0, 0, unix_io_manager, &fs))) {
 		errstr(_("error (%d) while opening %s\n"), (int)error, device);
@@ -429,22 +430,22 @@ static int ext2_direct_scan(const char *device)
 
 	if ((error = ext2fs_allocate_inode_bitmap(fs, "in-use inode map", &inode_used_map))) {
 		errstr(_("error (%d) while allocating file inode bitmap\n"), (int)error);
-		return -1;
+		goto out_free;
 	}
 
 	if ((error = ext2fs_allocate_inode_bitmap(fs, "directory inode map", &inode_dir_map))) {
 		errstr(_("errstr (%d) while allocating directory inode bitmap\n"), (int)error);
-		return -1;
+		goto out_used_map;
 	}
 
 	if ((error = ext2fs_open_inode_scan(fs, inode_buffer_blocks, &scan))) {
 		errstr(_("error (%d) while opening inode scan\n"), (int)error);
-		return -1;
+		goto out_dir_map;
 	}
 
 	if ((error = ext2fs_get_next_inode(scan, &i_num, &inode))) {
 		errstr(_("error (%d) while starting inode scan\n"), (int)error);
-		return -1;
+		goto out_scan;
 	}
 
 	while (i_num) {
@@ -474,10 +475,19 @@ static int ext2_direct_scan(const char *device)
 
 		if ((error = ext2fs_get_next_inode(scan, &i_num, &inode))) {
 			errstr(_("Something weird happened while scanning. Error %d\n"), (int)error);
-			return -1;
+			goto out_scan;
 		}
 	}
-	return 0;
+	ret = 0;
+out_scan:
+	ext2fs_close_inode_scan(scan);
+out_dir_map:
+	ext2fs_free_inode_bitmap(inode_dir_map);
+out_used_map:
+	ext2fs_free_inode_bitmap(inode_used_map);
+out_free:
+	ext2fs_free(fs);
+	return ret;
 }
 #endif
 
@@ -508,10 +518,15 @@ static int scan_dir(const char *pathname)
 		add_to_quota(GRPQUOTA, st.st_ino, st.st_uid, st.st_gid, st.st_mode,
 			     st.st_nlink, qspace, 0);
 
-	if ((dp = opendir(pathname)) == (DIR *) NULL)
-		die(2, _("\nCan open directory %s: %s\n"), pathname, strerror(errno));
+	if (chdir(pathname) == -1) {
+		errstr(_("Cannot chdir to %s: %s\n"), pathname, strerror(errno));
+		goto out;
+	}
+
+	if ((dp = opendir(".")) == (DIR *) NULL)
+		die(2, _("\nCannot open directory %s: %s\n"),
+		    pathname, strerror(errno));
 
-	chdir(pathname);
 	if (flags & FL_VERYVERBOSE)
 		blit(pathname);
 	while ((de = readdir(dp)) != (struct dirent *)NULL) {
@@ -595,8 +610,7 @@ int ask_yn(char *q, int def)
 
 	printf("%s [%c]: ", q, def ? 'y' : 'n');
 	fflush(stdout);
-	while (1) {
-		fgets(a, sizeof(a)-1, stdin);
+	while (fgets(a, sizeof(a)-1, stdin)) {
 		if (a[0] == '\n')
 			return def;
 		if (!strcasecmp(a, "y\n"))
@@ -606,6 +620,7 @@ int ask_yn(char *q, int def)
 		printf("Illegal answer. Please answer y/n: ");
 		fflush(stdout);
 	}
+	return def;
 }
 
 /* Do checks and buffer quota file into memory */
@@ -959,7 +974,10 @@ Please stop all programs writing to filesystem or use -m flag to force checking.
 start_scan:
 	debug(FL_VERBOSE | FL_DEBUG, _("Scanning %s [%s] "), mnt->me_devname, mnt->me_dir);
 #if defined(EXT2_DIRECT)
-	if (!strcmp(mnt->me_type, MNTTYPE_EXT2) || !strcmp(mnt->me_type, MNTTYPE_EXT3) || !strcmp(mnt->me_type, MNTTYPE_NEXT3)) {
+	if (!strcmp(mnt->me_type, MNTTYPE_EXT2) ||
+	    !strcmp(mnt->me_type, MNTTYPE_EXT3) ||
+	    !strcmp(mnt->me_type, MNTTYPE_NEXT3) ||
+	    !strcmp(mnt->me_type, MNTTYPE_EXT4)) {
 		if ((failed = ext2_direct_scan(mnt->me_devname)) < 0)
 			goto out;
 	}
