@@ -34,7 +34,9 @@
 
 #include "config.h"
 
+#if defined(RPC)
 #include <rpc/rpc.h>
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/file.h>
@@ -49,6 +51,8 @@
 #include <unistd.h>
 #include <time.h>
 #include <ctype.h>
+#include <limits.h>
+#include <stdlib.h>
 
 #if defined(RPC)
 #include "rquota.h"
@@ -95,7 +99,7 @@ struct dquot *getprivs(qid_t id, struct quota_handle **handles, int quiet)
 #if defined(BSD_BEHAVIOUR)
 	int j, ngroups;
 	uid_t euid;
-	gid_t gidset[NGROUPS], *gidsetp;
+	gid_t gidset[NGROUPS_MAX], *gidsetp;
 #endif
 
 	for (i = 0; handles[i]; i++) {
@@ -113,7 +117,7 @@ struct dquot *getprivs(qid_t id, struct quota_handle **handles, int quiet)
 				if (geteuid() == 0)
 					break;
 				ngroups = sysconf(_SC_NGROUPS_MAX);
-				if (ngroups > NGROUPS) {
+				if (ngroups > NGROUPS_MAX) {
 					gidsetp = malloc(ngroups * sizeof(gid_t));
 					if (!gidsetp) {
 						gid2group(id, name);
@@ -214,8 +218,10 @@ int editprivs(char *tmpfile)
 		int i;
 
 		sigprocmask(SIG_SETMASK, &omask, NULL);
-		setgid(getgid());
-		setuid(getuid());
+		if (setgid(getgid()))
+			die(1, _("%s failed: %s\n"), "setgid", strerror(errno));
+		if (setuid(getuid()))
+			die(1, _("%s failed: %s\n"), "setuid", strerror(errno));
 		if (!(ed = getenv("VISUAL")))
 			if (!(ed = getenv("EDITOR")))
 				ed = _PATH_VI;
@@ -245,6 +251,28 @@ int editprivs(char *tmpfile)
 	return 0;
 }
 
+/*
+ * Duplicate a file descriptor, resetting the file offset.
+ * If it's a write descriptor, also truncate the file.
+ */
+static FILE *dup_file(int fd, int is_write)
+{
+	FILE *fp;
+
+	if (is_write && ftruncate(fd, 0))
+		die(1, _("Cannot truncate a file: %s\n"), strerror(errno));
+	if (lseek(fd, 0, SEEK_SET))
+		die(1, _("Cannot reset a file offset: %s\n"), strerror(errno));
+	if ((fd = dup(fd)) < 0)
+		die(1, _("Cannot duplicate a file descriptor: %s\n"), strerror(errno));
+	if (!(fp = fdopen(fd, is_write ? "w" : "r")))
+		die(1, is_write ? _("Cannot open a stream to write to: %s\n")
+				: _("Cannot open a stream to read from: %s\n"),
+		    strerror(errno));
+
+	return fp;
+}
+
 /*
  * Convert a dquot list to an ASCII file.
  */
@@ -253,10 +281,7 @@ int writeprivs(struct dquot *qlist, int outfd, char *name, int quotatype)
 	struct dquot *q;
 	FILE *fd;
 
-	ftruncate(outfd, 0);
-	lseek(outfd, 0, SEEK_SET);
-	if (!(fd = fdopen(dup(outfd), "w")))
-		die(1, _("Cannot duplicate descriptor of file to write to: %s\n"), strerror(errno));
+	fd = dup_file(outfd, 1);
 
 	fprintf(fd, _("Disk quotas for %s %s (%cid %d):\n"),
 		_(type2name(quotatype)), name, *type2name(quotatype), qlist->dq_id);
@@ -317,15 +342,17 @@ int readprivs(struct dquot *qlist, int infd)
 	char inodesstring[BUFSIZ], isoftstring[BUFSIZ], ihardstring[BUFSIZ];
 	const char *error;
 
-	lseek(infd, 0, SEEK_SET);
-	if (!(fd = fdopen(dup(infd), "r")))
-		die(1, _("Cannot duplicate descriptor of temp file: %s\n"), strerror(errno));
+	fd = dup_file(infd, 0);
 
 	/*
 	 * Discard title lines, then read lines to process.
 	 */
-	fgets(line, sizeof(line), fd);
-	fgets(line, sizeof(line), fd);
+	if (!fgets(line, sizeof(line), fd) ||
+	    !fgets(line, sizeof(line), fd)) {
+		errstr(_("Bad format: two title lines assumed\n"));
+		fclose(fd);
+		return -1;
+	}
 
 	while (fgets(line, sizeof(line), fd)) {
 		cnt = sscanf(line, "%s %s %s %s %s %s %s",
@@ -425,10 +452,7 @@ int writeindividualtimes(struct dquot *qlist, int outfd, char *name, int quotaty
 	time_t now;
 	char btimestr[MAXTIMELEN], itimestr[MAXTIMELEN];
 
-	ftruncate(outfd, 0);
-	lseek(outfd, 0, SEEK_SET);
-	if (!(fd = fdopen(dup(outfd), "w")))
-		die(1, _("Cannot duplicate descriptor of file to write to: %s\n"), strerror(errno));
+	fd = dup_file(outfd, 1);
 
 	fprintf(fd, _("Times to enforce softlimit for %s %s (%cid %d):\n"),
 		_(type2name(quotatype)), name, *type2name(quotatype), qlist->dq_id);
@@ -468,16 +492,18 @@ int readindividualtimes(struct dquot *qlist, int infd)
 	char iunits[BUFSIZ], bunits[BUFSIZ];
 	time_t now, bseconds, iseconds;
 
-	lseek(infd, 0, SEEK_SET);
-	if (!(fd = fdopen(dup(infd), "r")))
-		die(1, _("Cannot duplicate descriptor of temp file: %s\n"), strerror(errno));
+	fd = dup_file(infd, 0);
 
 	/*
 	 * Discard title lines, then read lines to process.
 	 */
-	fgets(line, sizeof(line), fd);
-	fgets(line, sizeof(line), fd);
-	fgets(line, sizeof(line), fd);
+	if (!fgets(line, sizeof(line), fd) ||
+	    !fgets(line, sizeof(line), fd) ||
+	    !fgets(line, sizeof(line), fd)) {
+		errstr(_("Bad format: three title lines assumed\n"));
+		fclose(fd);
+		return -1;
+	}
 
 	time(&now);
 	while (fgets(line, sizeof(line), fd)) {
@@ -529,10 +555,7 @@ int writetimes(struct quota_handle **handles, int outfd)
 	if (!handles[0])
 		return 0;
 
-	ftruncate(outfd, 0);
-	lseek(outfd, 0, SEEK_SET);
-	if ((fd = fdopen(dup(outfd), "w")) == NULL)
-		die(1, _("Cannot duplicate descriptor of file to edit: %s\n"), strerror(errno));
+	fd = dup_file(outfd, 1);
 
 	fprintf(fd, _("Grace period before enforcing soft limits for %ss:\n"),
 		_(type2name(handles[0]->qh_type)));
@@ -561,12 +584,8 @@ int readtimes(struct quota_handle **handles, int infd)
 
 	if (!handles[0])
 		return 0;
-	lseek(infd, 0, SEEK_SET);
-	if (!(fd = fdopen(dup(infd), "r"))) {
-		errstr(_("Cannot reopen temp file: %s\n"),
-			strerror(errno));
-		return -1;
-	}
+
+	fd = dup_file(infd, 0);
 
 	/* Set all grace times to default values */
 	for (i = 0; handles[i]; i++) {
@@ -577,9 +596,13 @@ int readtimes(struct quota_handle **handles, int infd)
 	/*
 	 * Discard three title lines, then read lines to process.
 	 */
-	fgets(line, sizeof(line), fd);
-	fgets(line, sizeof(line), fd);
-	fgets(line, sizeof(line), fd);
+	if (!fgets(line, sizeof(line), fd) ||
+	    !fgets(line, sizeof(line), fd) ||
+	    !fgets(line, sizeof(line), fd)) {
+		errstr(_("Bad format: three title lines assumed\n"));
+		fclose(fd);
+		return -1;
+	}
 
 	while (fgets(line, sizeof(line), fd)) {
 		cnt = sscanf(line, "%s %d %s %d %s", fsp, &btime, bunits, &itime, iunits);
