@@ -79,6 +79,95 @@ char *type2name(int type)
 	return extensions[type];
 }
 
+/*
+ *	Project quota handling
+ */
+#define PROJECT_FILE "/etc/projid"
+#define MAX_LINE_LEN 1024
+
+static FILE *project_file;
+
+/* Rewind /etc/projid to the beginning */
+void setprent(void)
+{
+	if (project_file)
+		fclose(project_file);
+	project_file = fopen(PROJECT_FILE, "r");
+}
+
+/* Close /etc/projid file */
+void endprent(void)
+{
+	if (project_file) {
+		fclose(project_file);
+		project_file = NULL;
+	}
+}
+
+/* Get next entry in /etc/projid */
+struct fs_project *getprent(void)
+{
+	static struct fs_project p;
+	static char linebuf[MAX_LINE_LEN];
+	char *idstart, *idend;
+
+	if (!project_file)
+		return NULL;
+	while (fgets(linebuf, MAX_LINE_LEN, project_file)) {
+		/* Line too long? */
+		if (linebuf[strlen(linebuf) - 1] != '\n')
+			continue;
+		/* Skip comments */
+		if (linebuf[0] == '#')
+			continue;
+		idstart = strchr(linebuf, ':');
+		/* Skip invalid lines... We follow what xfs_quota does */
+		if (!idstart)
+			continue;
+		*idstart = 0;
+		idstart++;
+		/*
+		 * Colon can separate name from something else - follow what
+		 * xfs_quota does
+		 */
+		idend = strchr(idstart, ':');
+		if (idend)
+			*idend = 0;
+		p.pr_name = linebuf;
+		p.pr_id = strtoul(idstart, NULL, 10);
+		return &p;
+	}
+	return NULL;
+}
+
+static struct fs_project *get_project_by_name(char *name)
+{
+	struct fs_project *p;
+
+	setprent();
+	while ((p = getprent())) {
+		if (!strcmp(name, p->pr_name))
+			break;
+	}
+	endprent();
+
+	return p;
+}
+
+static struct fs_project *get_project_by_id(qid_t id)
+{
+	struct fs_project *p;
+
+	setprent();
+	while ((p = getprent())) {
+		if (p->pr_id == id)
+			break;
+	}
+	endprent();
+
+	return p;
+}
+
 /*
  *	Convert name to uid
  */
@@ -137,6 +226,36 @@ gid_t group2gid(char *name, int flag, int *err)
 	return entry->gr_gid;
 }
 
+/*
+ *	Convert project name to project id
+ */
+qid_t project2pid(char *name, int flag, int *err)
+{
+	qid_t ret;
+	char *errch;
+	struct fs_project *p;
+
+	if (err)
+		*err = 0;
+	if (!flag) {
+		ret = strtoul(name, &errch, 0);
+		if (!*errch)		/* Is name number - we got directly pid? */
+			return ret;
+	}
+	p = get_project_by_name(name);
+	if (!p) {
+		if (!err) {
+			errstr(_("project %s does not exist.\n"), name);
+			exit(1);
+		}
+		else {
+			*err = -1;
+			return 0;
+		}
+	}
+	return p->pr_id;
+}
+
 /*
  *	Convert name to id
  */
@@ -144,8 +263,10 @@ int name2id(char *name, int qtype, int flag, int *err)
 {
 	if (qtype == USRQUOTA)
 		return user2uid(name, flag, err);
-	else
+	else if (qtype == GRPQUOTA)
 		return group2gid(name, flag, err);
+	else
+		return project2pid(name, flag, err);
 }
 
 /*
@@ -180,6 +301,22 @@ int gid2group(gid_t id, char *buf)
 	return 0;
 }
 
+/*
+ *	Convert project id to name
+ */
+int pid2project(qid_t id, char *buf)
+{
+	struct fs_project *p;
+
+	if (!(p = get_project_by_id(id))) {
+		snprintf(buf, MAXNAMELEN, "#%u", (uint) id);
+		return 1;
+	}
+	else
+		sstrncpy(buf, p->pr_name, MAXNAMELEN);
+	return 0;
+}
+
 /*
  *	Convert id to user/groupname
  */
@@ -187,8 +324,10 @@ int id2name(int id, int qtype, char *buf)
 {
 	if (qtype == USRQUOTA)
 		return uid2user(id, buf);
-	else
+	else if (qtype == GRPQUOTA)
 		return gid2group(id, buf);
+	else
+		return pid2project(id, buf);
 }
 
 /*
@@ -531,6 +670,8 @@ static int hasxfsquota(const char *dev, struct mntent *mnt, int type, int flags)
 			return QF_XFS;
 		else if (type == GRPQUOTA && (info.qs_flags & XFS_QUOTA_GDQ_ACCT))
 			return QF_XFS;
+		else if (type == PRJQUOTA && (info.qs_flags & XFS_QUOTA_PDQ_ACCT))
+			return QF_XFS;
 #ifdef XFS_ROOTHACK
 		/*
 		 * Old XFS filesystems (up to XFS 1.2 / Linux 2.5.47) had a
@@ -543,6 +684,8 @@ static int hasxfsquota(const char *dev, struct mntent *mnt, int type, int flags)
 			return QF_XFS;
 		else if (type == GRPQUOTA && (sbflags & XFS_QUOTA_GDQ_ACCT))
 			return QF_XFS;
+		else if (type == PRJQUOTA && (sbflags & XFS_QUOTA_PDQ_ACCT))
+			return QF_XFS;
 #endif /* XFS_ROOTHACK */
 	}
 
@@ -882,18 +1025,11 @@ void init_kernel_interface(void)
 		die(2, _("Cannot set signal handler: %s\n"), strerror(errno));
 
 	kernel_qfmt_num = 0;
-	if (!stat("/proc/fs/xfs/stat", &st))
-		kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
-	else {
-		fs_quota_stat_t dummy;
-
-		if (!quotactl(QCMD(Q_XGETQSTAT, 0), "/dev/root", 0, (void *)&dummy) || (errno != EINVAL && errno != ENOSYS))
-			kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
-	}
 	/* Detect new kernel interface; Assume generic interface unless we can prove there is not one... */
 	if (!stat("/proc/sys/fs/quota", &st) || errno != ENOENT) {
 		kernel_iface = IFACE_GENERIC;
 		kernel_qfmt[kernel_qfmt_num++] = QF_META;
+		kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
 		kernel_qfmt[kernel_qfmt_num++] = QF_VFSOLD;
 		kernel_qfmt[kernel_qfmt_num++] = QF_VFSV0;
 		kernel_qfmt[kernel_qfmt_num++] = QF_VFSV1;
@@ -901,6 +1037,15 @@ void init_kernel_interface(void)
 	else {
 		struct v2_dqstats v2_stats;
 
+		if (!stat("/proc/fs/xfs/stat", &st))
+			kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
+		else {
+			fs_quota_stat_t dummy;
+
+			if (!quotactl(QCMD(Q_XGETQSTAT, 0), "/dev/root", 0, (void *)&dummy) ||
+			    (errno != EINVAL && errno != ENOSYS))
+				kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
+		}
 		if (quotactl(QCMD(Q_V2_GETSTATS, 0), NULL, 0, (void *)&v2_stats) >= 0) {
 			kernel_qfmt[kernel_qfmt_num++] = QF_VFSV0;
 			kernel_iface = IFACE_VFSV0;
@@ -971,18 +1116,34 @@ static int v2_kern_quota_on(const char *dev, int type)
 	return 0;
 }
 
-/* Check whether XFS quota is turned on on given device */
-static int xfs_kern_quota_on(const char *dev, int type)
+/*
+ * Check whether quota is turned on on given device. This quotactl always
+ * worked for XFS and it works even for VFS quotas for kernel 4.1 and newer.
+ *
+ * We return 0 when quota is not turned on, 1 when only accounting is turned
+ * on, and 2 when both accounting and enforcement is turned on. We return -1
+ * on error.
+ */
+int kern_quota_state_xfs(const char *dev, int type)
 {
 	struct xfs_mem_dqinfo info;
 
 	if (!quotactl(QCMD(Q_XFS_GETQSTAT, type), dev, 0, (void *)&info)) {
-		if (type == USRQUOTA && (info.qs_flags & XFS_QUOTA_UDQ_ACCT))
-			return 1;
-		if (type == GRPQUOTA && (info.qs_flags & XFS_QUOTA_GDQ_ACCT))
-			return 1;
+		if (type == USRQUOTA) {
+			return !!(info.qs_flags & XFS_QUOTA_UDQ_ACCT) +
+			       !!(info.qs_flags & XFS_QUOTA_UDQ_ENFD);
+		}
+		if (type == GRPQUOTA) {
+			return !!(info.qs_flags & XFS_QUOTA_GDQ_ACCT) +
+			       !!(info.qs_flags & XFS_QUOTA_GDQ_ENFD);
+		}
+		if (type == PRJQUOTA) {
+			return !!(info.qs_flags & XFS_QUOTA_PDQ_ACCT) +
+			       !!(info.qs_flags & XFS_QUOTA_PDQ_ENFD);
+		}
+		return 0;
 	}
-	return 0;
+	return -1;
 }
 
 /*
@@ -996,7 +1157,7 @@ int kern_quota_on(struct mount_entry *mnt, int type, int fmt)
 		return -1;
 	if (mnt->me_qfmt[type] == QF_XFS) {
 		if ((fmt == -1 || fmt == QF_XFS) &&
-	    	    xfs_kern_quota_on(mnt->me_devname, type))	/* XFS quota format */
+		    kern_quota_state_xfs(mnt->me_devname, type) > 0)
 			return QF_XFS;
 		return -1;
 	}
@@ -1137,12 +1298,13 @@ alloc:
 			free((char *)devname);
 			devname = sstrdup(loopdev);
 		}
-
 		/* Further we are not interested in mountpoints without quotas and
 		   we don't want to touch them */
 		qfmt[USRQUOTA] = hasquota(devname, mnt, USRQUOTA, flags);
 		qfmt[GRPQUOTA] = hasquota(devname, mnt, GRPQUOTA, flags);
-		if (qfmt[USRQUOTA] < 0 && qfmt[GRPQUOTA] < 0) {
+		qfmt[PRJQUOTA] = hasquota(devname, mnt, PRJQUOTA, flags);
+		if (qfmt[USRQUOTA] < 0 && qfmt[GRPQUOTA] < 0 &&
+		    qfmt[PRJQUOTA] < 0) {
 			free((char *)devname);
 			continue;
 		}
